/*****************************************************************************************[Cooperation.h]
 Copyright (c) 2008-2011, Youssef Hamadi, Saïd Jabbour and Lakhdar Saïs
 
 Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
 associated documentation files (the "Software"), to deal in the Software without restriction,
 including without limitation the rights to use, copy, modify, merge, publish, distribute,
 sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in all copies or
 substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT
 NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
 DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT
 OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *******************************************************************************************/
#include "core/Bess.h"

#include<cassert>
#include<cmath>

namespace Minisat {

/*_________________________________________________________________________________________________
 |
 |  getRealBandits : (id)  ->  [int*]
 |  Description: returns the real bandits.
 |________________________________________________________________________________________________@*/

int* Bess::getRealBandits(int id) {

	int *rbandits = new int[nbBandits];

	int i, j;
	for (i = j = 0; i < nbThreads; i++)
		if (BanditM[id][i] == 1)
			rbandits[j++] = i;

	return rbandits;
}

/*_________________________________________________________________________________________________
 |
 |  getRealBandits : (id)  ->  [int*]
 |  Description: returns the real bandits.
 |________________________________________________________________________________________________@*/

void Bess::getSleptBandits(int id) {

	int i, j;
	for (i = j = 0; i < nbThreads; i++)
		if (BanditM[id][i] == 0 && id != i)
			sbandits[id][j++] = i;

}

/*_________________________________________________________________________________________________
 |
 |  getSleepBandits : (id)  ->  [int]
 |  Description: returns the first free bandit.
 |________________________________________________________________________________________________@*/

int Bess::getSleepBandit(int id) {

	//assert(indexFbandit[id]>-1 && indexFbandit[id]<nbThreads-nbBandits-1);
	int fbandit = sbandits[id][indexFbandit[id]];
	indexFbandit[id] = (indexFbandit[id] + 1) % (nbThreads - nbBandits - 1);

	return fbandit;
}

/*_________________________________________________________________________________________________
 |
 |  swapBandit : (id, banditOut[nb])  ->  [int]
 |  Description : swap the set of banditOut from awake to sleep state
 |________________________________________________________________________________________________@*/

int Bess::swapBandit(int id, int* banditOut) {

	if (nbThreads - nbBandits == -1)
		return 0;
	int fbandit;
	for (int t = 0; t < nbBandits; t++) {

		if (banditOut[t] != -1) {
			BanditM[id][banditOut[t]] = 0;
			fbandit = getSleepBandit(id);
			BanditM[id][fbandit] = 1;
			//print:
			/*if (id == 0) {
			 for (int i = 0; i < nbThreads - nbBandits - 1; i++)
			 printf("%2d ", sbandits[0][i]);
			 printf("     => (%d,%d)\n", fbandit, banditOut[t]);
			 }*/

			if (indexFbandit[id] == 0)
				sbandits[id][(nbThreads - nbBandits) - 2] = banditOut[t];

			else
				sbandits[id][indexFbandit[id] - 1] = banditOut[t];
			//	current_bandit[id][t] = fbandit;


			//print
			/*if (id == 0) {
			 for (int i = 0; i < (nbThreads - nbBandits) - 1; i++)
			 printf("%2d ", sbandits[0][i]);
			 printf("\n\n");
			 }*/

		}
	}

	return 0;
}

/*_________________________________________________________________________________________________
 |
 |  isBanditOf : (int id , int i)  ->  [bool]
 |  Description : checks if i is a real bandit fo id
 |________________________________________________________________________________________________@*/

bool Bess::isBanditOf(int id, int i) {
	return (BanditM[id][i] == 1);
}

/*_________________________________________________________________________________________________
 |
 |  randomBandit : (id)  ->  [void]
 |  Description : swap bandits randomly
 |________________________________________________________________________________________________@*/

void Bess::randomBandits(int id) {

	int ncores[nbThreads];
	int rbandit, temp;

	for (int t = 0; t < nbThreads; t++)
		ncores[t] = t;

	for (int t = 0; t < nbThreads; t++) {
		rbandit = rand() % nbThreads;
		temp = ncores[t];
		ncores[t] = ncores[rbandit];
		ncores[rbandit] = temp;
	}

	for (int t = 0; t < nbThreads; t++)
		BanditM[id][t] = 0;

	for (int t = 0; t < nbBandits; t++)
		BanditM[id][ncores[t]] = 1;

	if (BanditM[id][id] == 1) {
		BanditM[id][id] = 0;
		BanditM[id][ncores[nbBandits]] = 1;
	}

}
/*_________________________________________________________________________________________________
 |
 |  randomOneBandit : (id)  ->  [void]
 |  Description : swap one bandit randomly
 |________________________________________________________________________________________________@*/

void Bess::randomOneBandit(int id) {

	int rbandit;

	for (int t = 0; t < nbThreads; t++) {
		rbandit = rand() % nbThreads;
		if (BanditM[id][t] + BanditM[id][rbandit] == 1) {
			BanditM[id][rbandit] = (BanditM[id][rbandit] + 3) % 2;
			BanditM[id][t] = (BanditM[id][t] + 3) % 2;
			assert(BanditM[id][t] + BanditM[id][rbandit] == 1);

			break;
		}
	}

}

/*_________________________________________________________________________________________________
 |
 |  manageDBaseBandits : (id)  ->  [void]
 |  Description : manage the bandits state (awake, sleep)
 |________________________________________________________________________________________________@*/

void Bess::manageDBaseBandits(int id) {

	bwindow[id] = 0;

	double current_reward = 0;

	int banditOut[nbBandits]; //  for mode0 and mode1
	for (int t = 0; t < nbBandits; t++)
		banditOut[t] = -1;

	//BEGIN------------------------------- reward VALUE for each real bandits---------------------------------------------

	Reward_function(id);


	if (!mutex[id]) {
		mutex[id] = true;
		rewardThreshold[id] /= (double) nbBandits;
	}

	//END------------------------------- reward VALUE for each real bandits---------------------------------------------


	/*******************************************************************
	 * 						BANDITs SWAPPING 						   *
	 * 																   *
	 *******************************************************************/

	switch (MODE) {
	/*-----------------------  ONE BANDIT UNDER THRESHOLD (mode0) ---------------------------*/



		/*-----------------------  ALL BANDIT UNDER THRESHOLD (mode1)---------------------------*/
	case 1: {
		int ind = 0; //  BanditOut index

		for (int t = 0; t < nbThreads; t++)
			if (isBanditOf(id, t) && (BanditReward[id][t]
					<= rewardThreshold[id])) {

				if (t != id) {
					banditOut[ind] = t;
					ind++;
				}

			}

		swapBandit(id, banditOut);

		break;
	}

		/*-----------------------  ONE RANDOM SWAPPING (mode2) ---------------------------*/
	case 2: {
		randomOneBandit(id);

		break;
	}

		/*-----------------------  ALL RANDOM SWAPPING (mode3) ---------------------------*/

	case 3: {
		randomBandits(id);

		break;
	}
		/*-----------------------          NO SWAP (mode4)        ---------------------------*/
	case 4: {
		break;
	}

	}

	/*
	 * Threashold
	 */
	switch (METHOD) {
	case 0: {
		current_reward = 0;
		for (int t = 0; t < nbThreads; t++)
			if (isBanditOf(id, t)) {
				current_reward += BanditReward[id][t];
			}
		rewardThreshold[id] = .9 * rewardThreshold[id] + .1
				* ((double) current_reward / (double) nbBandits);
		break;
	}
	case 1: { //without threshold calculus for random modes
		break;

	}
	}

	// if(id==0) printBanditM();
}



/*_________________________________________________________________________________________________
 |
 |  Reward_function : (id)  ->  [void]
 |  Description : a reward function switch the setin mode:

 |________________________________________________________________________________________________@*/

void Bess:: Reward_function(int id){

for (int t = 0; t < nbThreads; t++) {
	if (isBanditOf(id, t)) {

		for (int i = 0; i < limitExportClauses; i++) {

			switch (Reward) {

			case C_size: { //RECEIVED/SIZE clauses

				BanditReward[id][t] += (double) -(received[id][t][i]
						* log10(1 - (pow(2, -(i + 1))))); //LOG

				break;
			}
			case C_vsids: { //VSIDS QUALITY BASED


				BanditReward[id][t] += -((double) (kept[id][t][i]
						/ (received[id][t][i] + 1.)) * 100. * log10(1
						- (pow(2, -(i + 1)))));

				break;
			}
			case C_vsids_continu: { //LBD


				/*if (id == 0)
				 printf("size=%d  quantity=%d lbd=%d reward=%g\n",
				 i + 1, received[id][t][i], kept[id][t][i],
				 BanditReward[id][t]);*/
				break;
			}

			}

		}

		//  THRESHOLD INITIALIZATION

		if (!mutex[id] || rewardThreshold[id] == 0)
			rewardThreshold[id] += (double) BanditReward[id][t];

	}

}
}

/*_________________________________________________________________________________________________
 |
 |  initPercentage : (received, percentage)  ->  [void]
 |  Description : initialize BanditReward, received and percentage matrix
 |________________________________________________________________________________________________@*/

void Bess::initReceived(double** BanditReward, int*** received,
		int*** kept) {

	for (int t1 = 0; t1 < nbThreads; t1++)
		for (int t2 = 0; t2 < nbThreads; t2++) {
			BanditReward[t1][t2] = 0;

			for (int i = 0; i < limitExportClauses; i++) {
				received[t1][t2][i] = 0;
				kept[t1][t2][i] = 0;

			}

		}

}


void Bess::printReward(int t, double** Reward, int ind) {
	char filename[30];

	FILE* LOG;

	sprintf(filename, "./LOG/core%d_%d.log", t, ind);

	LOG = fopen(filename, "a");

	if (isBanditOf(0, t)) {
		fprintf(LOG, "%g \n", Reward[0][t]);

	}

	fclose(LOG);

}


/*_________________________________________________________________________________________________
 |
 |  PrintBandits : ()  ->  [void]
 |  Description: print the set of real bandits
 |________________________________________________________________________________________________@*/

void Bess::printBandits(int id) {

	char filename[30];
	FILE* LOG;

	sprintf(filename, "./LOG/bandit_core%d.log", id);

	LOG = fopen(filename, "a");
	for (int t = 0; t < nbThreads; t++)
		if (isBanditOf(id, t))
			fprintf(LOG, "%d ", t);

	fprintf(LOG, "\n");
	fclose(LOG);

}
void Bess::printBanditM() {
	FILE* LOG;

	LOG = fopen("./LOG/banditM.log", "a");
	fprintf(LOG, "----------------------------------\n");
	for (int i = 0; i < nbThreads; i++) {
		for (int j = 0; j < nbThreads; j++)
			fprintf(LOG, "%2d ", BanditM[i][j]);
		fprintf(LOG, "\n");
	}
	fprintf(LOG, "\n----------------------------------\n");

	fclose(LOG);
}
/*_________________________________________________________________________________________________
 |
 |  printTotalClauseR : ()  ->  [void]
 |  Description: print the number of received clauses in each generation
 |________________________________________________________________________________________________@*/

void Bess::printTotalClauseR(int id, int*** received) {

	char filename[30];
	FILE* LOG;

	sprintf(filename, "./LOG/totalClauseR%d.log", id);

	LOG = fopen(filename, "a");
	int sum = 0;
	for (int t = 0; t < nbThreads; t++)
		for (int i = 0; i < limitExportClauses; i++)
			if (isBanditOf(id, t))
				sum += received[id][t][i];

	fprintf(LOG, "%d\n", sum);
	fclose(LOG);

}

/*_________________________________________________________________________________________________
 |
 |  printClausePerSize : ()  ->  [void]
 |  Description: print the number of received clauses in each generation
 |________________________________________________________________________________________________@*/

void Bess::printClausePerSize(int id, int*** received) {

	char filename[30];
	for (int i = 0; i < limitExportClauses; i++) {
		FILE* LOG;
		sprintf(filename, "./LOG/Core%dSize%d.log", id, i + 1);
		LOG = fopen(filename, "a");

		for (int t = 0; t < nbThreads; t++)

		{
			if (isBanditOf(id, t))
				fprintf(LOG, "%d ", received[id][t][i]);

		}
		fprintf(LOG, "\n");
		fclose(LOG);
	}

}

bool Bess::inTab(int* tab, int size, int val) {
	for (int t = 0; t < size; t++)
		if (tab[t] == val)
			return true;
	return false;

}

void Bess::printNorm(int t, double** Reward, int ind) {
	char filenameN[30];
	FILE* LOGN;

	double moyenneR = 0;
	double varianceR = 0;
	double normalR = 0;
	for (int t = 1; t < nbThreads; t++) {
		moyenneR += Reward[0][t];
	}
	moyenneR /= (double) (nbThreads - 1);

	for (int t = 1; t < nbThreads; t++) {

		varianceR += pow(Reward[0][t] - moyenneR, 2);
	}
	varianceR /= (double) (nbThreads - 1);

	for (int t = 1; t < nbThreads; t++) {
		sprintf(filenameN, "./LOG/norm%d_%d.log", t, ind);
		LOGN = fopen(filenameN, "a");

		if (varianceR != 0)
			normalR = (Reward[0][t] - moyenneR) / varianceR;
		else
			normalR = 0;
		fprintf(LOGN, "%g\n", normalR);

		fclose(LOGN);
	}

}

}
