/*****************************************************************************************[Cooperation.h]
 Copyright (c) 2008-20011, Youssef Hamadi, Saïd Jabbour and Lakhdar Saïs
 
 Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
 associated documentation files (the "Software"), to deal in the Software without restriction,
 including without limitation the rights to use, copy, modify, merge, publish, distribute,
 sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in all copies or
 substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT
 NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
 DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT
 OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *******************************************************************************************/

#include "core/SolverTypes.h"
#include "core/Solver.h"

namespace Minisat {

//=================================================================================================
// Options:

#define MODE  1
/*
 * mode = 0 : one bandit swapped
 * mode = 1 : all bandits < Threashold swapped
 * mode = 2 : One bandit random
 * mode = 3 : All bandits random
 * mode = 4 : no swap
 */

// Reward functions
#define C_size			1		// -log(1-1/2^size)
#define C_lbd			2		// C_size/lbd
#define C_size_lbd  	3		// -log(1-1/2^lbd)
#define C_vsids			4		// Q * C_size  ( Q VSIDS based quality)
#define C_vsids_continu	5			// f( sum f(p(literal)) / nb(literaux)
//f(x) = exp (alpha (1 - 1/x))
//p(literal) = activite(literal) / activite_max

#define METHOD 0
/*
 * method = 0 : weighted threshold method
 * method = 1 : without threshold calculus for random modes
 */

/*=================================================================================================
 Cooperation Class : ->  [class]
 Description:
 This class manage clause sharing component between threads,i.e.,
 It controls read and write operations in extraUnits and extraClauses arrays
 iterators headExtra... and tailExtra... are used to to this end
 =================================================================================================*/

class Bess {

public:

	//BANDIT SECTION
	int nbBandits;
	int nbThreads;
	int Reward;
	int limitExportClauses;
	int* bwindow;
	int* window;
	int** BanditM;
	double** BanditReward;
	int*** received;
	int*** kept;

	bool* mutex;
	double* rewardThreshold;
	int* indexFbandit;
	int** sbandits; // slept bandits matrix

	double random_seed;
	//int ** nbSizeLearnt;

	//=================================================================================================

	int swapBandit(int numThread, int* banditOut);
	void randomBandits(int id);
	void randomOneBandit(int id);
	bool isBanditOf(int id, int i);
	void getSleptBandits(int id);
	int getSleepBandit(int id);
	void manageDBaseBandits(int numThread);
	void Reward_function(int id);
	int* getRealBandits(int numThread);
	void initReceived(double** BanditReward, int*** received, int*** kept);

	void printBandits(int id);
	void printTotalClauseR(int id, int*** received);
	void printClausePerSize(int id, int*** received);
	void printBanditM();
	bool inTab(int* tab, int size, int val);
	void printReward(int t, double** BanditReward, int ind);
	void printNorm(int t, double** BanditReward, int ind);

	Bess(int n, int b, int l) {

		nbThreads=n;
		nbBandits=b;
		limitExportClauses=l;
		BanditM = new int*[nbThreads];
		BanditReward = new double*[nbThreads];
		received = new int**[nbThreads];
		kept = new int**[nbThreads];
		bwindow = new int[nbThreads];
		window = new int[nbThreads];
		mutex = new bool[nbThreads];
		rewardThreshold = new double[nbThreads];
		indexFbandit = new int[nbThreads];
		sbandits = new int*[nbThreads];

		for (int t = 0; t < nbThreads; t++) {

			//Bandit
			BanditM[t] = new int[nbThreads];
			BanditReward[t] = new double[nbThreads];
			received[t] = new int*[nbThreads];
			kept[t] = new int*[nbThreads];

			bwindow[t] = 0;
			window[t] = 0;
			mutex[t] = false;
			rewardThreshold[t] = 0;
			indexFbandit[t] = 0;
			//nbSizeLearnt[t] = new int[8];
			sbandits[t] = new int[nbThreads - nbBandits - 1];

			for (int k = 0; k < nbThreads; k++) {

				received[t][k] = new int[l];
				kept[t][k] = new int[l];

				if (k < nbBandits)
					BanditM[t][(t + 1 + k) % nbThreads] = 1;

				BanditReward[t][k] = 0;

				for (int i = 0; i < l; i++) {
					received[t][k][i] = 0;
					kept[t][k][i] = 0;
				}

			}

			getSleptBandits(t);

			/*	for (int k = 0; k < 8; k++)
			 nbSizeLearnt[t][k] = 0;*/
		}
		for (int i = 0; i < nbThreads; i++)
			for (int j = 0; j < nbThreads; j++)
				if (BanditM[i][j] != 1)
					BanditM[i][j] = 0;
		for (int i = 0; i < nbThreads; i++) {
			for (int j = 0; j < nbThreads; j++)
				printf("%2d ", BanditM[i][j]);
			printf("\n");
		}

		printf("Slept Bandits:\n");
		for (int i = 0; i < nbThreads; i++) {
			for (int j = 0; j < nbThreads - nbBandits - 1; j++)
				printf("%2d ", sbandits[i][j]);
			printf("\n");
		}

	}
	//=================================================================================================
	~Bess() {
	}
};
}

