/*****************************************************************************************[Cooperation.h]
 Copyright (c) 2008-2011, Youssef Hamadi, Saïd Jabbour and Lakhdar Saïs
 
 Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
 associated documentation files (the "Software"), to deal in the Software without restriction,
 including without limitation the rights to use, copy, modify, merge, publish, distribute,
 sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in all copies or
 substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT
 NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
 DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT
 OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *******************************************************************************************/
#include "core/Cooperation.h"
#include<cassert>
#include<cmath>

namespace Minisat {

/*_________________________________________________________________________________________________
 |
 |  getAwakeBandits : (id)  ->  [int*]
 |  Description: returns the real bandits.
 |________________________________________________________________________________________________@*/
/*
 int* Cooperation::getAwakeBandits(int id) {

 int *rbandits = new int[nbBandits];

 int i, j;
 for (i = j = 0; i < nbThreads; i++)
 if (BanditM[id][i] == 1)
 rbandits[j++] = i;

 return rbandits;
 }
 */
/*_________________________________________________________________________________________________
 |
 |  getSleptBandits : (id)  ->  [int*]
 |  Description: returns the slept bandits.
 |________________________________________________________________________________________________@*/

void Cooperation::getSleptBandits(int id) {

	int i, j;
	for (i = j = 0; i < nbThreads; i++)
		if (BanditM[id][i] == 0 && id != i)
			sbandits[id][j++] = i;

}

/*_________________________________________________________________________________________________
 |
 |  getSleepBandit : (id)  ->  [int]
 |  Description: returns the first free bandit.
 |________________________________________________________________________________________________@*/

int Cooperation::getSleepBandit(int id) {

	//assert(indexSbandit[id]>-1 && indexSbandit[id]<nbThreads-nbBandits-1);
	int fbandit = sbandits[id][indexSbandit[id]];
	indexSbandit[id] = (indexSbandit[id] + 1) % (nbThreads - nbBandits - 1);

	return fbandit;
}

/*_________________________________________________________________________________________________
 |
 |  swapBandit : (id, banditOut[nb])  ->  [int]
 |  Description : swap the set of banditOut from awake to sleep state
 |________________________________________________________________________________________________@*/

void Cooperation::swapBandit(int id, int* banditOut) {

	if (nbThreads - nbBandits != 1) {
		int fbandit;
		for (int t = 0; t < nbBandits; t++) {

			if (banditOut[t] != -1) {
				BanditM[id][banditOut[t]] = 0;
				BanditReward[id][banditOut[t]] = 0; // RAZ reward of sleeping bandit!!
				fbandit = getSleepBandit(id);
				BanditM[id][fbandit] = 1;
				Flag[id][fbandit] = true;
				//print:
				/*if (id == 0) {
				 for (int i = 0; i < nbThreads - nbBandits - 1; i++)
				 printf("%2d ", sbandits[0][i]);
				 printf("     => (%d,%d)\n", fbandit, banditOut[t]);
				 }*/

				if (indexSbandit[id] == 0)
					sbandits[id][(nbThreads - nbBandits) - 2] = banditOut[t];

				else
					sbandits[id][indexSbandit[id] - 1] = banditOut[t];
				//	current_bandit[id][t] = fbandit;


				//print
				/*if (id == 0) {
				 for (int i = 0; i < (nbThreads - nbBandits) - 1; i++)
				 printf("%2d ", sbandits[0][i]);
				 printf("\n\n");
				 }*/

			}
		}
	}

}

/*_________________________________________________________________________________________________
 |
 |  isBanditOf : (int id , int i)  ->  [bool]
 |  Description : checks if i is a real bandit fo id
 |________________________________________________________________________________________________@*/

bool Cooperation::isBanditOf(int id, int i) {
	return (BanditM[id][i] == 1);
}

/*_________________________________________________________________________________________________
 |
 |  randomBandit : (id)  ->  [void]
 |  Description : swap bandits randomly
 |________________________________________________________________________________________________@*/

void Cooperation::randomSwap(int id) {

	int ncores[nbThreads];
	int rbandit, temp;

	for (int t = 0; t < nbThreads; t++)
		ncores[t] = t;

	for (int t = 0; t < nbThreads; t++) {
		rbandit = rand() % nbThreads;
		temp = ncores[t];
		ncores[t] = ncores[rbandit];
		ncores[rbandit] = temp;
	}

	for (int t = 0; t < nbThreads; t++)
		BanditM[id][t] = 0;

	for (int t = 0; t < nbBandits; t++)
		BanditM[id][ncores[t]] = 1;

	if (BanditM[id][id] == 1) {
		BanditM[id][id] = 0;
		BanditM[id][ncores[nbBandits]] = 1;
	}

}
/*_________________________________________________________________________________________________
 |
 |  oneRandomSwap : (id)  ->  [void]
 |  Description : swap one bandit randomly
 |________________________________________________________________________________________________@*/

void Cooperation::oneRandomSwap(int id) {

	int rbandit = rand() % nbThreads;

	for (int t = 0; t < nbThreads && rbandit == id; t++)
		rbandit = rand() % nbThreads;

	for (int t = 0; t < nbThreads; t++) {

		if (BanditM[id][t] + BanditM[id][rbandit] == 1 && id != t && id
				!= rbandit) {
			BanditM[id][rbandit] = (BanditM[id][rbandit] + 3) % 2;
			BanditM[id][t] = (BanditM[id][t] + 3) % 2;
			assert(BanditM[id][t] + BanditM[id][rbandit] == 1);
			printf("one random swap DONE !!\n");
			break;
		}
	}
	printf("exit\n");
}

/*_________________________________________________________________________________________________
 |
 |  manageDBaseBandits : (id)  ->  [void]
 |  Description : manage the bandits state (awake, sleep)
 |________________________________________________________________________________________________@*/

void Cooperation::manageDBaseBandits(int id) {

	bwindow[id] = 0;
	luby_seq[id]++;

	int banditOut[nbBandits]; //  for threshold based swap
	for (int t = 0; t < nbBandits; t++)
		banditOut[t] = -1;

	//BEGIN------------------------------- reward VALUE for each awake bandit---------------------------------------------

	Reward_function(id);

	//END------------------------------- reward VALUE for each awake bandit---------------------------------------------

	if (!mutex[id]) {
		mutex[id] = true;
		rewardThreshold[id] /= (double) nbBandits;
	}

	/*******************************************************************
	 * 						BANDITs SWAPPING 						   *
	 * 																   *
	 *******************************************************************/

	switch (mode) {
	/*-----------------------  ONE BANDIT UNDER THRESHOLD (mode0) ---------------------------*/

	case one_swap: {
		double rewardMin = rewardThreshold[id];
		int bandit = id;
		int BanditWithZreward[nbBandits]; // for bandits with 0 reward, we select randomly on to be swapped
		int ind = 0;
		bool mutex = false;

		for (int t = 0; t < nbThreads; t++)
			if (isBanditOf(id, t) && BanditReward[id][t] <= rewardMin) {
				if (BanditReward[id][t] == 0) {
					BanditWithZreward[ind] = t;
					ind++;
					mutex = true;
				}
				rewardMin = BanditReward[id][t];
				bandit = t;
			}

		if (mutex) {
			int rand = irand(random_seed, (ind), 0);
			bandit = BanditWithZreward[rand];
		}
		banditOut[0] = bandit;

		swapBandit(id, banditOut);

		break;
	}

		/*-----------------------  ALL BANDIT UNDER THRESHOLD (mode1)---------------------------*/
	case all_swap: {
		int ind = 0; //  BanditOut index

		for (int t = 0; t < nbThreads; t++)
			if (isBanditOf(id, t) && (BanditReward[id][t]
					<= rewardThreshold[id])) {

				if (t != id) {
					banditOut[ind] = t;
					ind++;
				}

			}

		swapBandit(id, banditOut);

		break;
	}

		/*-----------------------  ONE RANDOM SWAPPING (mode2) ---------------------------*/
	case one_random_swap: {
		oneRandomSwap(id);

		break;
	}

		/*-----------------------  ALL RANDOM SWAPPING (mode3) ---------------------------*/

	case all_random_swap: {
		randomSwap(id);

		break;
	}
		/*-----------------------          NO SWAP (mode4)        ---------------------------*/
	case no_swap: {
		break;
	}

	}

	/*
	 * Threashold update
	 */
	if (mode == one_swap || mode == all_swap)
		update_threshold(id);
}

/*_________________________________________________________________________________________________
 |
 |  update_threshold : (id)  ->  [void]
 |  Description : weighted threshold update using a relaxation factor alpha=.1

 |________________________________________________________________________________________________@*/

void Cooperation::update_threshold(int id) {

	double current_reward = 0;
	int counter = 0;
	for (int t = 0; t < nbThreads; t++)
		if (isBanditOf(id, t) && Flag[id][t]) {
			current_reward += RewardInstant[id][t];
			Flag[id][t] = false;
			counter++;
		}
	if (counter != 0)
		rewardThreshold[id] = .9 * rewardThreshold[id] + .1
				* ((double) current_reward / (double) counter);

	// if(id==0) printf("%g\n",rewardThreshold[0]);
}

/*_________________________________________________________________________________________________
 |
 |  Reward_function : (id)  ->  [void]
 |  Description : a reward function switch the setin mode:

 |________________________________________________________________________________________________@*/

void Cooperation::Reward_function(int id) {

	for (int t = 0; t < nbThreads; t++) {
		if (isBanditOf(id, t)) {

			for (int i = 0; i < limitExportClauses; i++) {

				switch (Reward) {

				case C_size: { //RECEIVED/SIZE clauses

					RewardInstant[id][t] += (double) -(received[id][t][i]
							* log10(1 - (pow(2, -(i + 1))))); //LOG

					break;
				}
				case C_vsids: { //VSIDS QUALITY BASED


					RewardInstant[id][t] += -((double) (kept[id][t][i]
							/ (received[id][t][i] + 1.)) * 100. * log10(1
							- (pow(2, -(i + 1)))));

					break;
				}

				}

			}

			// Average Cumulative reward with relaxation factor alpha=.1

			if (!mutex[id])
				BanditReward[id][t] = RewardInstant[id][t];
			else
				BanditReward[id][t] = (.9 * BanditReward[id][t]) + (.1
						* RewardInstant[id][t]);

			//  THRESHOLD INITIALIZATION

			if (!mutex[id] || rewardThreshold[id] == 0)
				rewardThreshold[id] += (double) BanditReward[id][t];

		}
	}

}

/*_________________________________________________________________________________________________
 |
 |  initPercentage : (received, percentage)  ->  [void]
 |  Description : initialize RewardInstant, received and percentage matrix
 |________________________________________________________________________________________________@*/

void Cooperation::initReward(double** RewardInstant, int*** received,
		int*** kept) {

	for (int t1 = 0; t1 < nbThreads; t1++)
		for (int t2 = 0; t2 < nbThreads; t2++) {
			RewardInstant[t1][t2] = 0;

			for (int i = 0; i < limitExportClauses; i++) {
				received[t1][t2][i] = 0;
				kept[t1][t2][i] = 0;

			}

		}

}

/*_________________________________________________________________________________________________
 |
 |  exportExtraUnit : ()  ->  [void]
 |  Description : manage export Extra Unit Clauses
 |________________________________________________________________________________________________@*/

void Cooperation::exportExtraUnit(Solver* s, Lit unit) {

	int id = s->threadId;

	for (int t = 0; t < nbThreads; t++) {

		if (BanditM[t][id] == 0)
			continue;
		if (t == id)
			continue;

		int ind = tailExtraUnits[id][t];
		if (((ind + 1) % MAX_EXTRA_UNITS) == headExtraUnits[id][t])
			continue;

		extraUnits[id][t][ind++].x = unit.x;
		if (ind == MAX_EXTRA_UNITS)
			ind = 0;
		tailExtraUnits[id][t] = ind;
	}
}

/*_________________________________________________________________________________________________
 |
 |  importExtraUnits : ()  ->  [void]
 |  Description : store unit extra clause in the vector unit_learnts
 |________________________________________________________________________________________________@*/

void Cooperation::importExtraUnits(Solver* s, vec<Lit>& unit_learnts) {

	int id = s->threadId;

	for (int t = 0; t < nbThreads; t++) {

		if (BanditM[id][t] == 0)
			continue;

		if (t == id)
			continue;

		int head = headExtraUnits[t][id];
		int tail = tailExtraUnits[t][id];
		if (head == tail)
			continue;

		int end = tail;
		if (tail < head)
			end = MAX_EXTRA_UNITS;

		for (int i = head; i < end; i++)
			storeExtraUnits(s, t, extraUnits[t][id][i], unit_learnts);

		if (tail < head)
			for (int i = 0; i < tail; i++)
				storeExtraUnits(s, t, extraUnits[t][id][i], unit_learnts);

		head = tail;
		if (head == MAX_EXTRA_UNITS)
			head = 0;
		headExtraUnits[t][id] = head;
	}
}

/*_________________________________________________________________________________________________
 |
 |  importExtraUnits : ()  ->  [void]
 |  Description : manage import Extra Unit Clauses
 |________________________________________________________________________________________________@*/

void Cooperation::importExtraUnits(Solver* s) {

	int id = s->threadId;

	for (int t = 0; t < nbThreads; t++) {

		if (BanditM[id][t] == 0)
			continue;

		if (t == id)
			continue;

		int head = headExtraUnits[t][id];
		int tail = tailExtraUnits[t][id];
		if (head == tail)
			continue;

		int end = tail;
		if (tail < head)
			end = MAX_EXTRA_UNITS;

		for (int i = head; i < end; i++)
			uncheckedEnqueue(s, t, extraUnits[t][id][i]);

		if (tail < head)
			for (int i = 0; i < tail; i++)
				uncheckedEnqueue(s, t, extraUnits[t][id][i]);

		head = tail;
		if (head == MAX_EXTRA_UNITS)
			head = 0;
		headExtraUnits[t][id] = head;
	}
}

/*_________________________________________________________________________________________________
 |
 |  uncheckedEnqueue : ()  ->  [void]
 |  Description: Enqueue the unit literals at level 0
 |________________________________________________________________________________________________@*/

void Cooperation::uncheckedEnqueue(Solver* s, int t, Lit l) {

	if (s->value(l) == l_False)
		answers[s->threadId] = l_False;

	if (s->value(l) != l_Undef)
		return;
	s->uncheckedEnqueue(l);
	nbImportedExtraUnits[s->threadId]++;

}

/*_________________________________________________________________________________________________
 |
 |  uncheckedEnqueue : ()  ->  [void]
 |  Description: Enqueue the unit literals at level 0
 |________________________________________________________________________________________________@*/

void Cooperation::storeExtraUnits(Solver* s, int t, Lit l, vec<Lit>& unitLits) {
	int id = s->threadId;
	unitLits.push(l);
	nbImportedExtraUnits[id]++;

	//BANDIT
	received[id][t][0]++;

	kept[id][t][0]++;

	if (Reward == C_lbd || Reward == C_size_lbd)
		RewardInstant[id][t] += 0.301029995664; //    (-(log10(1 - (pow(2, -(1))))));
	if (Reward == C_vsids_continu)
		RewardInstant[id][t] ++;
}

/*_________________________________________________________________________________________________
 |
 |  exportExtraClause : ()  ->  [void]
 |  Description: each time a new clause is learnt, it is exported to other threads
 |________________________________________________________________________________________________@*/

void Cooperation::exportExtraClause(Solver* s, vec<Lit>& learnt) {

	int id = s->threadId;

	for (int t = 0; t < nbThreads; t++) {

		assert(BanditM[t][id]==0 || BanditM[t][id]==1);
		assert(id >= 0 && id <= nbThreads);
		if (t == id)
			continue;
		else if (BanditM[t][id] == 0)
			continue; // && (learnt.size() != 2)
		if (learnt.size() > PrwLimitExportClauses[id][t])
			continue;

		int ind = tailExtraClauses[id][t];
		if (((ind + 1) % MAX_EXTRA_CLAUSES) == headExtraClauses[id][t])
			continue;

		extraClauses[id][t][ind] = new Lit[learnt.size() + 1];
		extraClauses[id][t][ind][0] = mkLit(learnt.size());

		for (int j = 0; j < learnt.size(); j++)
			extraClauses[id][t][ind][j + 1] = learnt[j];

		ind++;
		if (ind == MAX_EXTRA_CLAUSES)
			ind = 0;
		tailExtraClauses[id][t] = ind;
	}
}

/*_________________________________________________________________________________________________
 |
 |  exportExtraClause : ()  ->  [void]
 |  Description: each time a new clause is learnt, it is exported to other threads
 |________________________________________________________________________________________________@*/

void Cooperation::exportExtraClause(Solver* s, Clause& c) {

	int id = s->threadId;

	for (int t = 0; t < nbThreads; t++) {

		assert(BanditM[t][id]==0 || BanditM[t][id]==1);
		if (t == id)
			continue;
		else if (BanditM[t][id] == 0)
			continue; //&& (c.size() != 2))
		if (c.size() > PrwLimitExportClauses[id][t])
			continue;

		//   if((BanditM [t][id] == 0)  || (c.size() > pairwiseLimitExportClauses[id][t]))
		//continue;

		int ind = tailExtraClauses[id][t];
		if (((ind + 1) % MAX_EXTRA_CLAUSES) == headExtraClauses[id][t])
			continue;

		extraClauses[id][t][ind] = new Lit[c.size() + 1];
		extraClauses[id][t][ind][0] = mkLit(c.size());

		for (int j = 0; j < c.size(); j++)
			extraClauses[id][t][ind][j + 1] = c[j];

		ind++;
		if (ind == MAX_EXTRA_CLAUSES)
			ind = 0;
		tailExtraClauses[id][t] = ind;
	}
}

/*_________________________________________________________________________________________________
 |
 |  importExtraClauses : ()  ->  [void]
 |  Description: import Extra clauses sent from other threads
 |_______________________________________________________________________________________________@*/

void Cooperation::importExtraClauses(Solver* s) {

	int id = s->threadId;

	for (int t = 0; t < nbThreads; t++) {

		if (t == id)
			continue;

		if (BanditM[id][t] == 0)
			continue;

		int head = headExtraClauses[t][id];
		int tail = tailExtraClauses[t][id];
		if (head == tail)
			continue;

		int end = tail;
		if (tail < head)
			end = MAX_EXTRA_CLAUSES;

		for (int i = head; i < end; i++)
			addExtraClause(s, t, extraClauses[t][id][i]);

		if (tail < head)
			for (int i = 0; i < tail; i++)
				addExtraClause(s, t, extraClauses[t][id][i]);

		head = tail;
		if (head == MAX_EXTRA_CLAUSES)
			head = 0;
		headExtraClauses[t][id] = head;
	}
}

/*_________________________________________________________________________________________________
 |
 |  addExtraClause : ()  ->  [void]
 |  Description: build a clause from the learnt Extra Lit*
 |  watch it correctly, test basic cases distinguich other cases during watching process
 |________________________________________________________________________________________________@*/

void Cooperation::addExtraClause(Solver* s, int t, Lit* lt) {

	vec<Lit> extra_clause;
	int extra_backtrack_level = 0;
	int id = s->threadId;
	int size = var(lt[0]);
	int wtch = 0;
	//	Lit q = lit_Undef;
	int lbd = 0;

	for (int i = 1, j = 0; i < size + 1; i++) {
		Lit q = lt[i];
		if (s->value(q) == l_False && s->level(var(q)) == 0)
			continue;
		extra_clause.push(q);
		if (s->value(q) != l_False && wtch < 2) {
			extra_clause[j] = extra_clause[wtch];
			extra_clause[wtch++] = q;
		} else if (s->level(var(q)) >= extra_backtrack_level)
			extra_backtrack_level = s->level(var(q));
		j++;
	}

	//conflict clause at level 0 --> formula is UNSAT
	if (extra_clause.size() == 0)
		setAnswer(id, l_False);
	if (extra_clause.size() > 10000)
		printf("SIZE=%d\n", extra_clause.size());

	// LBD measure for the received extra-clause

	if (Reward == C_lbd || Reward == C_size_lbd) { // lbd based reward!!
		// Find the LBD measure
		lbd = extra_clause.size();
		int diff_level[extra_clause.size()]; // level of each literal
		bool mutex = false; //if mutex=T: all literals have 0 as level=> lbd=size

		for (int i = 0; i < extra_clause.size(); i++) {
			diff_level[i] = s->level(var(extra_clause[i]));
			if (!mutex && diff_level[i] == 0)
				mutex = true;
		}

		for (int i = 0; i < extra_clause.size() - 1; i++)
			for (int j = i + 1; j < extra_clause.size(); j++)
				if (diff_level[i] != 0 && diff_level[i] == diff_level[j]) {
					lbd--;
					break;
				}
		if (mutex && lbd == 1)
			lbd = extra_clause.size();

		if (Reward == C_lbd && lbd != 0)
			RewardInstant[id][t] += (-(log10(1
					- (pow(2, -(extra_clause.size()))))) / lbd);

		if (Reward == C_size_lbd && lbd != 0)
			RewardInstant[id][t] += (-log10(1 - (pow(2, -(lbd)))));

	}
	//------------------------------------------


	// case of unit extra clause
	else if (extra_clause.size() == 1) {
		s->cancelUntil(0);
		if (s->value(extra_clause[0]) == l_Undef) {
			s->uncheckedEnqueue(extra_clause[0]);
			CRef cs = s->propagate();
			if (cs != CRef_Undef)
				answers[id] = l_False;
		}
	}

	else {
		// build clause from lits and add it to learnts(s) base
		CRef cr = s->addExtraClause(extra_clause);

		// Case of Unit propagation: literal to propagate or bad level
		if (wtch == 1 && (s->value(extra_clause[0]) == l_Undef
				|| extra_backtrack_level < s->level(var(extra_clause[0])))) {
			s->cancelUntil(extra_backtrack_level);
			s->uncheckedEnqueue(extra_clause[0], cr);
		}

		// Case of Conflicting Extra Clause --> analyze that conflict
		else if (wtch == 0) {
			extra_clause.clear();
			s->cancelUntil(extra_backtrack_level);

			s->analyze(cr, extra_clause, extra_backtrack_level);
			s->cancelUntil(extra_backtrack_level);

			// analyze lead to unit clause
			if (extra_clause.size() == 1) {
				assert(extra_backtrack_level == 0);
				if (s->value(extra_clause[0]) == l_Undef)
					s->uncheckedEnqueue(extra_clause[0]);
			}
			// analyze lead to clause with size > 1
			else {
				CRef cs = s->addExtraClause(extra_clause);
				s->uncheckedEnqueue(extra_clause[0], cs);
				//exportExtraClause(s, extra_clause);
			}
		}
	}

	nbImportedExtraClauses[id]++;

	//VSIDS based reward!!
	if (Reward == C_vsids) {
		int nbactiv = 0;

		for (int j = 1; j < size + 1; j++) {
			if ((s->activity[var(lt[j])] > s->max_act / 2))
				nbactiv++;
		}
		if (size != 0 && (nbactiv >= (double) size / 3))
			kept[id][t][size - 1]++;
	}

	if (Reward == C_vsids_continu) {

		double s_continu = 0;

		for (int j = 1; j < size + 1; j++)
			s_continu += f_continu(s->activity[var(lt[j])] / s->max_act);

		if (size != 0)
			s_continu /= size;
		RewardInstant[id][t] += f_continu(s_continu);

	}
	if (size != 0)
		received[id][t][size - 1]++;

	delete[] lt;
}

/*_________________________________________________________________________________________________
 |
 |  f_continu : (double)  ->  [double]
 |  Description:
 |  		f(x)= [ (1/(1+e(-alpha(2x-1))) - beta ] * gamma
 |________________________________________________________________________________________________@*/

double Cooperation::f_continu(double x) {

	if (x == 0)
		return 0;
	if (x == 1)
		return 1;

	assert(((1 / (1 + exp(-ALPHA * (2 * x - 1))) - BETA) * GAMMA)>0 &&((1 / (1 + exp(-ALPHA * (2 * x - 1))) - BETA) * GAMMA)<1);
	return ((1 / (1 + exp(-ALPHA * (2 * x - 1))) - BETA) * GAMMA); // ALPHA = 5,
	//BETA = 1/1+exp(ALPHA) = 0.006692
	//GAMMA=(2+exp(-ALPHA)+exp(ALPHA))/exp(ALPHA)-exp(-ALPHA) = 1.013529

}

/*_________________________________________________________________________________________________
 |
 |  addExtraClause : ()  ->  [void]
 |  Description: watch lits correctly and testing basic cases and
 |  build a clause from the learnt Extra Lit*
 |________________________________________________________________________________________________@*/
/*
 void Cooperation::addExtraClause1(Solver* s, int t, Lit* lt) {

 int id = s->threadId;
 int size = var(lt[0]);

 vec<Lit> lits;
 int j = 0;
 int k = 0;
 //	Lit q = lit_Undef;
 int maxLevel = 0;

 for (int i = 1; i < size + 1; i++) {
 Lit q = lt[i];
 if (s->value(q) == l_False && s->level(var(q)) == 0)
 continue;
 lits.push(lt[i]);
 if (s->value(q) != l_False) {
 lits[k] = lits[j];
 lits[j++] = q;
 } else if (s->level(var(q)) >= maxLevel)
 maxLevel = s->level(var(q));
 k++;
 }

 //conflict clause at level 0
 if (lits.size() == 0)
 setAnswer(id, l_False);

 // case of unit extra clause
 else if (lits.size() == 1) {
 s->cancelUntil(0);
 if (s->value(lits[0]) == l_Undef)
 s->uncheckedEnqueue(lits[0]);
 } else {
 // build clause from lits
 s->addExtraClause(lits);
 }

 nbImportedExtraClauses[id]++;

 delete[] lt;
 }
 */
/*_________________________________________________________________________________________________
 |
 |  addExtraClause : ()  ->  [void]
 |  Description: build a clause from the learnt Extra Lit*
 |________________________________________________________________________________________________@*/
/*
 void Cooperation::addExtraClause2(Solver* s, int t, Lit* lt) {

 int id = s->threadId;
 int size = var(lt[0]);
 vec<Lit> lits;

 for (int i = 1; i < size + 1; i++)
 lits.push(lt[i]);

 // build clause from lits
 s->addExtraClause(lits);

 nbImportedExtraClauses[id]++;

 delete[] lt;
 }
 */

/*_________________________________________________________________________________________________
 |
 |  printStats : ()  ->  [void]
 |  Description: print statistics of each thread
 |________________________________________________________________________________________________@*/

void Cooperation::printStats(int& id) {

	uint64_t nbSharedExtraClauses = 0;
	uint64_t nbSharedExtraUnits = 0;
	uint64_t totalConflicts = 0;

	for (int t = 0; t < nbThreads; t++) {

		nbSharedExtraClauses += nbImportedExtraClauses[t];
		nbSharedExtraUnits += nbImportedExtraUnits[t];
		totalConflicts += solvers[t].conflicts;
	}

	printf(
			" -----------------------------------------------------------------------------------------------------------------------\n");
	printf(
			"| ManySAT 2.2 - all threads statistics           DETERMINISTIC MODE ? %s              initial limit export clauses : %3d |\n",
			deterministic_mode == true ? "Y" : "N", limitExportClauses);
	printf(
			"|-----------------------------------------------------------------------------------------------------------------------|\n");
	printf(
			"|  Thread  | winner  |   #restarts   |   decisions   |  #conflicts   |   %% shared cls  |  #extra units | #extra clauses |         \n");
	printf(
			"|----------|---------|---------------|---------------|---------------|-----------------|---------------|----------------|\n");

	for (int t = 0; t < nbThreads; t++)
		printf("| %8d | %7s | %13d | %13d | %13d | %13d %% | %13d | %14d |\n",
				t, t == id ? "x" : " ", (int) solvers[t].starts,
				(int) solvers[t].decisions, (int) solvers[t].conflicts,
				(int) solvers[t].conflicts == 0 ? 0
						: (int) (nbImportedExtraUnits[t]
								+ nbImportedExtraClauses[t]) * 100
								/ (int) solvers[t].conflicts,
				(int) nbImportedExtraUnits[t], (int) nbImportedExtraClauses[t]);

	printf(
			"|----------------------------------------------------|---------------|-----------------|---------------|----------------|\n");
	printf(
			"|                                                    | %13d | %13d %% | %13d | %14d |\n",
			(int) totalConflicts, (int) totalConflicts == 0 ? 0
					: (int) (nbSharedExtraUnits + nbSharedExtraClauses) * 100
							/ (int) totalConflicts, (int) nbSharedExtraUnits,
			(int) nbSharedExtraClauses);
	printf(
			" -----------------------------------------------------------------------------------------------------------------------\n");

	Parallel_Info();

}

/*_________________________________________________________________________________________________
 |
 |  printExMatrix : ()  ->  [void]
 |  Description: print the final values of size clauses shared between parwise threads
 |________________________________________________________________________________________________@*/

void Cooperation::printExMatrix() {

	printf("\n Final Matrix extra shared clauses limit size \n");
	printf(" ----------------------------------------------------\n");
	for (int i = 0; i < nbThreads; i++) {
		printf("|");
		for (int j = 0; j < nbThreads; j++)
			if (i != j)
				printf(" e(%d,%d)=%3d  ", i, j,
						(int) PrwLimitExportClauses[i][j]);
			else
				printf(" e(%d,%d)=%3d  ", i, j, 0);
		printf(" |\n");
	}
	printf(" ----------------------------------------------------\n\n");
}

/*_________________________________________________________________________________________________
 |
 |  Parallel_Info : ()  ->  [void]
 |  Description: more informations about deterministic or non deterministic mode and final matrix print
 |________________________________________________________________________________________________@*/

void Cooperation::Parallel_Info() {
	printf(" DETERMINISTIC_MODE? : %s \n", (deterministic_mode == true ? "YES"
			: "NO"));
	printf(" CONTROL POLICY?	 : %s  \n",
			(ctrl != 0 ? (ctrl == 1 ? "DYNAMIC (INCREMENTAL  en= +- 1)"
					: "DYNAMIC (AIMD: en=  en - a x en, en + b/en)")
					: "STATIC (fixed Limit Export Clauses)"));
	if (ctrl == 0)
		printf(" LIMIT EXPORT CLAUSES (PAIRWISE) : [identity] x %d\n\n",
				limitExportClauses);
	else
		printExMatrix();
}

/*_________________________________________________________________________________________________
 |
 |  PrintBandits : ()  ->  [void]
 |  Description: print the set of real bandits
 |________________________________________________________________________________________________@*/

void Cooperation::printBandits(int id) {

	char filename[30];
	FILE* LOG;

	sprintf(filename, "./LOG/bandit_core%d.log", id);

	LOG = fopen(filename, "a");
	for (int t = 0; t < nbThreads; t++)
		if (isBanditOf(id, t))
			fprintf(LOG, "%d ", t);

	fprintf(LOG, "\n");
	fclose(LOG);

}
void Cooperation::printBanditM() {
	FILE* LOG;

	LOG = fopen("./LOG/banditM.log", "a");
	fprintf(LOG, "----------------------------------\n");
	for (int i = 0; i < nbThreads; i++) {
		for (int j = 0; j < nbThreads; j++)
			fprintf(LOG, "%2d ", BanditM[i][j]);
		fprintf(LOG, "\n");
	}
	fprintf(LOG, "\n----------------------------------\n");

	fclose(LOG);
}
/*_________________________________________________________________________________________________
 |
 |  printTotalClauseR : ()  ->  [void]
 |  Description: print the number of received clauses in each generation
 |________________________________________________________________________________________________@*/

void Cooperation::printTotalClauseR(int id, int*** received) {

	char filename[30];
	FILE* LOG;

	sprintf(filename, "./LOG/totalClauseR%d.log", id);

	LOG = fopen(filename, "a");
	int sum = 0;
	for (int t = 0; t < nbThreads; t++)
		for (int i = 0; i < limitExportClauses; i++)
			if (isBanditOf(id, t))
				sum += received[id][t][i];

	fprintf(LOG, "%d\n", sum);
	fclose(LOG);

}


/*
 Finite subsequences of the Luby-sequence:

 0: 1
 1: 1 1 2
 2: 1 1 2 1 1 2 4
 3: 1 1 2 1 1 2 4 1 1 2 1 1 2 4 8
 ...


 */

 double Cooperation::luby(double y, int x) {

	// Find the finite subsequence that contains index 'x', and the
	// size of that subsequence:
	int size, seq;
	for (size = 1, seq = 0; size < x + 1; seq++, size = 2 * size + 1)
		;

	while (size - 1 != x) {
		size = (size - 1) >> 1;
		seq--;
		x = x % size;
	}

	return pow(y, seq);
}
/*_________________________________________________________________________________________________
 |
 |  printClausePerSize : ()  ->  [void]
 |  Description: print the number of received clauses in each generation
 |________________________________________________________________________________________________@*/

void Cooperation::printClausePerSize(int id, int*** received) {

	char filename[30];
	for (int i = 0; i < limitExportClauses; i++) {
		FILE* LOG;
		sprintf(filename, "./LOG/Core%dSize%d.log", id, i + 1);
		LOG = fopen(filename, "a");

		for (int t = 0; t < nbThreads; t++)

		{
			if (isBanditOf(id, t))
				fprintf(LOG, "%d ", received[id][t][i]);

		}
		fprintf(LOG, "\n");
		fclose(LOG);
	}

}

bool Cooperation::inTab(int* tab, int size, int val) {
	for (int t = 0; t < size; t++)
		if (tab[t] == val)
			return true;
	return false;

}

void Cooperation::printReward(int t, double** Reward, int ind) {
	char filename[30];

	FILE* LOG;

	sprintf(filename, "./LOG/core%d_%d.log", t, ind);

	LOG = fopen(filename, "a");

	if (isBanditOf(0, t)) {
		fprintf(LOG, "%g \n", Reward[0][t]);

	}

	fclose(LOG);

}

void Cooperation::printNorm(int t, double** Reward, int ind) {
	char filenameN[30];
	FILE* LOGN;

	double moyenneR = 0;
	double varianceR = 0;
	double normalR = 0;
	for (int t = 1; t < nbThreads; t++) {
		moyenneR += Reward[0][t];
	}
	moyenneR /= (double) (nbThreads - 1);

	for (int t = 1; t < nbThreads; t++) {

		varianceR += pow(Reward[0][t] - moyenneR, 2);
	}
	varianceR /= (double) (nbThreads - 1);

	for (int t = 1; t < nbThreads; t++) {
		sprintf(filenameN, "./LOG/norm%d_%d.log", t, ind);
		LOGN = fopen(filenameN, "a");

		if (varianceR != 0)
			normalR = (Reward[0][t] - moyenneR) / varianceR;
		else
			normalR = 0;
		fprintf(LOGN, "%g\n", normalR);

		fclose(LOGN);
	}

}

}
