/*****************************************************************************************[Cooperation.h]
 Copyright (c) 2008-20011, Youssef Hamadi, Saïd Jabbour and Lakhdar Saïs
 
 Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
 associated documentation files (the "Software"), to deal in the Software without restriction,
 including without limitation the rights to use, copy, modify, merge, publish, distribute,
 sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in all copies or
 substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT
 NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
 DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT
 OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *******************************************************************************************/

#include "core/SolverTypes.h"
#include "core/Solver.h"

namespace Minisat {

//=================================================================================================
// Options:

#define MAX_EXTRA_CLAUSES     2000
#define MAX_EXTRA_UNITS       2000
#define MAX_IMPORT_CLAUSES    4000
#define LIMIT_CONFLICTS_EVAL  6000
#define INITIAL_DET_FREQUENCE  700
#define EXTRA_LIMIT_SIZE        41     //max size of extra clause
//#define WINDOW_BANDIT_UPDATE  5000

// Reward functions

#define C_size			1		// -log(1-1/2^size)		DONE
#define C_lbd			2		// C_size/lbd			DONE
#define C_size_lbd  	3		// -log(1-1/2^lbd)		DONE
#define C_vsids			4		// Q * C_size  ( Q VSIDS based quality)		DONE
#define C_vsids_continu	5			// f( sum f(p(literal)) / nb(literaux)	DONE
									//f(x) = (1/(1+e(-alpha(2x-1))) - beta) * gamma
									//p(literal) = activite(literal) / activite_max

#define ALPHA			5			// VSIDS_continu parameters
#define BETA			0.006692	//   1/1+exp(ALPHA)
#define GAMMA			1.013529	// (2+exp(-ALPHA)+exp(ALPHA))/exp(ALPHA)-exp(-ALPHA)



#define one_swap		0		// one bandit under threshold swapped
#define all_swap		1		// all bandits under threshold swapped
#define one_random_swap	2		// one bandit randomly swapped
#define all_random_swap	3		// all bandits randomly swapped
#define no_swap			4		// no swap !!





/*=================================================================================================
 Cooperation Class : ->  [class]
 Description:
 This class manage clause sharing component between threads,i.e.,
 It controls read and write operations in extraUnits and extraClauses arrays
 iterators headExtra... and tailExtra... are used to to this end
 =================================================================================================*/

class Cooperation {

public:

	int nbThreads; // numbe of  threads
	int limitExportClauses; // initial limit size of shared clauses
	bool start, end; // start and end multi-threads search

	lbool* answers; // answer of threads
	Solver* solvers; // set of running CDCL algorithms
	double** PrwLimitExportClauses; // pairwised limit limit size clause sharing


	int** headExtraClauses;
	int** tailExtraClauses;
	int** headExtraUnits;
	int** tailExtraUnits;

	Lit*** extraUnits; // where are stored the shared unit clauses
	Lit**** extraClauses; // where are stored the set of shared clauses with size > 1


	int initFreq; // barrier synchronization limit conflicts in deterministic case
	char ctrl; // activate control clause sharing size mode
	bool deterministic_mode; // running Minisat in deterministic mode
	int* deterministic_freq; // in dynamic case, #conflicts barrier synchronization
	int* nbImportedExtraUnits; // counters fot the total imported Unit clauses
	int* nbImportedExtraClauses; // counters fot the total imported Extra clauses
	//uint64_t* learntsz;


	//BANDIT SECTION
	int nbBandits;  // number of awake bandits
	int Reward;		// reward function (Main option)
	int mode;		// swapping mode (Main option)
	int gen;		// size of the generation
	int* bwindow;	//  bandit swap time window
	int* luby_seq;	// counter for the luby sequence
	//int* window;
	int** BanditM;	// bandits matrix  0: sleep, 1: awake
	double** RewardInstant;  // reward at instant T
	double** BanditReward;  // average cumulative reward with relaxation factor alpha=.1
	int*** received;		// nb of received k-sized clause by core_i
	int*** kept;			// nb of kept clause based on vsids quality

	bool* mutex;			// mutex for the first threshold calculus
	double* rewardThreshold;	// weighted threshold value
	int* indexSbandit;			// next slept bandit
	int** sbandits; // slept bandits matrix
	bool** Flag;		// flag for the previous swapped bandit from sleep to alive (used in threshold update)


	//int ** nbSizeLearnt;

	//=================================================================================================

	void exportExtraUnit(Solver* s, Lit unit);
	void importExtraUnits(Solver* s);
	void importExtraUnits(Solver* s, vec<Lit>& lits);

	void exportExtraClause(Solver* s, vec<Lit>& learnt);
	void exportExtraClause(Solver* s, Clause& c);
	void importExtraClauses(Solver* s);

	void addExtraClause(Solver* s, int t, Lit* lt);
	void addExtraClause1(Solver* s, int t, Lit* lt);
	void addExtraClause2(Solver* s, int t, Lit* lt);

	void uncheckedEnqueue(Solver* s, int t, Lit l);
	void storeExtraUnits(Solver* s, int t, Lit l, vec<Lit>& lits);

	//---------------------------------------
	//void updateLimitExportClauses(Solver* s);

	void printStats(int& id);
	void printExMatrix();
	void Parallel_Info();
	double random_seed;

	void swapBandit(int numThread, int* banditOut);		// swap the awake bandits of "banditOut" to sleep state and wake others
	void randomSwap(int id);							// a random swap on all awake bandits
	void oneRandomSwap(int id);							// a random swap on one awake bandit
	bool isBanditOf(int id, int i);						// is "i" an awake bandit of "id" ?
	void getSleptBandits(int id);						// return all slept bandits
	int getSleepBandit(int id);							// return the first slept bandit
	void manageDBaseBandits(int numThread);				// manage bandits at each time window
	void Reward_function(int id);						// reward functions
	void update_threshold(int id);						// threshold update
//	int* getAwakeBandits(int numThread);					// returns the awake bandits
	void initReward(double** RewardInstant, int*** received, int*** kept);  //intitialization for the next time window (generation)
	double f_continu(double x);							// a continu fct for vsids reward calculus
	double luby(double y, int x);

// stat print:

	void printBandits(int id);
	void printTotalClauseR(int id, int*** received);
	void printClausePerSize(int id, int*** received);
	void printBanditM();
	bool inTab(int* tab, int size, int val);
	void printReward(int t, double** RewardInstant, int ind);
	void printNorm(int t, double** RewardInstant, int ind);

	// Returns a random float 0 <= x < 1. Seed must never be 0.
	static inline double drand(double& seed, int threadId) {
		seed /= (1 + threadId);
		seed *= 1389796;
		int q = (int) (seed / 2147483647);
		seed -= (double) q * 2147483647;
		return seed / 2147483647;
	}

	// Returns a random integer 0 <= x < size. Seed must never be 0.
	static inline int irand(double& seed, int size, int threadId) {
		return (int) (drand(seed, threadId) * size);
	}

	//=================================================================================================
	inline int nThreads() {
		return nbThreads;
	}
	inline int limitszClauses() {
		return limitExportClauses;
	}
	inline lbool answer(int t) {
		return answers[t];
	}
	inline bool setAnswer(int id, lbool lb) {
		answers[id] = lb;
		return true;
	}

	//=================================================================================================
	// Constructor / Destructor

	Cooperation(int n, int l, int b) {

		limitExportClauses = l;
		nbThreads = n;
		nbBandits = b;
		end = false;
		solvers = new Solver[nbThreads];
		answers = new lbool[nbThreads];

		extraUnits = new Lit**[nbThreads];
		headExtraUnits = new int*[nbThreads];
		tailExtraUnits = new int*[nbThreads];

		extraClauses = new Lit***[nbThreads];
		headExtraClauses = new int*[nbThreads];
		tailExtraClauses = new int*[nbThreads];

		BanditM = new int*[nbThreads];
		RewardInstant = new double*[nbThreads];
		BanditReward = new double*[nbThreads];
		received = new int**[nbThreads];
		kept = new int**[nbThreads];
		bwindow = new int[nbThreads];
		luby_seq= new int[nbThreads];
		//window = new int[nbThreads];
		mutex = new bool[nbThreads];
		rewardThreshold = new double[nbThreads];
		indexSbandit = new int[nbThreads];
		sbandits = new int*[nbThreads];
		Flag= new bool*[nbThreads];

		//nbSizeLearnt = new int*[nbThreads];
		for (int t = 0; t < nbThreads; t++) {
			extraUnits[t] = new Lit*[nbThreads];
			headExtraUnits[t] = new int[nbThreads];
			tailExtraUnits[t] = new int[nbThreads];

			extraClauses[t] = new Lit**[nbThreads];
			headExtraClauses[t] = new int[nbThreads];
			tailExtraClauses[t] = new int[nbThreads];

			//Bandit
			BanditM[t] = new int[nbThreads];
			RewardInstant[t] = new double[nbThreads];
			BanditReward[t] = new double[nbThreads];
			received[t] = new int*[nbThreads];
			kept[t] = new int*[nbThreads];
			Flag[t]= new bool[nbThreads];

			bwindow[t] = 0;
			luby_seq[t] = 0;

			//window[t] = 0;
			mutex[t] = false;
			rewardThreshold[t] = 0;
			indexSbandit[t] = 0;
			//nbSizeLearnt[t] = new int[8];
			sbandits[t] = new int[nbThreads - nbBandits - 1];

			for (int k = 0; k < nbThreads; k++) {
				extraUnits[t][k] = new Lit[MAX_EXTRA_UNITS];
				headExtraUnits[t][k] = 0;
				tailExtraUnits[t][k] = 0;

				extraClauses[t][k] = new Lit*[MAX_EXTRA_CLAUSES];
				headExtraClauses[t][k] = 0;
				tailExtraClauses[t][k] = 0;

				received[t][k] = new int[l];
				kept[t][k] = new int[l];

				Flag[t][k]=false;

				if (k < nbBandits)
					BanditM[t][(t + 1 + k) % nbThreads] = 1;

				RewardInstant[t][k] = 0;
				BanditReward[t][k] = 0;

				for (int i = 0; i < l; i++) {
					received[t][k][i] = 0;
					kept[t][k][i] = 0;
				}

			}

			getSleptBandits(t);

			/*	for (int k = 0; k < 8; k++)
			 nbSizeLearnt[t][k] = 0;*/
		}
		for (int i = 0; i < nbThreads; i++)
			for (int j = 0; j < nbThreads; j++)
				if (BanditM[i][j] != 1)
					BanditM[i][j] = 0;
		for (int i = 0; i < nbThreads; i++) {
			for (int j = 0; j < nbThreads; j++)
				printf("%2d ", BanditM[i][j]);
			printf("\n");
		}

		printf("Slept Bandits:\n");
		for (int i = 0; i < nbThreads; i++) {
			for (int j = 0; j < nbThreads - nbBandits - 1; j++)
				printf("%2d ", sbandits[i][j]);
			printf("\n");
		}
		/*exit(1);*/

		//=================================================================================================
		initFreq = INITIAL_DET_FREQUENCE;
		deterministic_mode = false;

		//learntsz = new uint64_t[nbThreads];
		deterministic_freq = new int[nbThreads];
		nbImportedExtraClauses = new int[nbThreads];
		nbImportedExtraUnits = new int[nbThreads];

		PrwLimitExportClauses = new double*[nbThreads];

		for (int t = 0; t < nbThreads; t++) {
			//learntsz[t] = 0;
			answers[t] = l_Undef;
			deterministic_freq[t] = initFreq;
			nbImportedExtraClauses[t] = 0;
			nbImportedExtraUnits[t] = 0;


			PrwLimitExportClauses[t] = new double[nbThreads];

			for (int k = 0; k < nbThreads; k++)
				PrwLimitExportClauses[t][k] = l;

		}
	}
	//=================================================================================================
	~Cooperation() {
	}
};
}

